"""
Constants related to the orientation and structure of the ellipsoid
"""
from mpmath import elliprd


def get_Ax(*, x, y, z):
    """
    Returns the degenerate Carlson symmetric elliptic integral of the third kind. Note that the permutation of the first
    two arguments does not change the result
    """
    return float(2/3 * (elliprd(y**2, z**2, x**2)))


def get_Ay(*, x, y, z):
    """
    Returns the degenerate Carlson symmetric elliptic integral of the third kind. Note that the permutation of the first
    two arguments does not change the result
    """
    return float(2/3 * (elliprd(x**2, z**2, y**2)))


def get_Az(*, x, y, z):
    """
    Returns the degenerate Carlson symmetric elliptic integral of the third kind. Note that the permutation of the first
    two arguments does not change the result
    """
    return float(2/3 * (elliprd(x**2, y**2, z**2)))





=======
"""
Functions and classes that can be useful.
"""
from enum import IntEnum
from math import sqrt, pi
from mpmath import findroot, mp
from numpy import linspace, arange
from scipy.optimize import brenth

from src.solve.geometry import get_Ax, get_Ay, get_Az


class ODEIndex(IntEnum):
    """
    Enumerated number for array index for variables in the ODEs
    """
    # Dimensionful units
    a1 = 0
    a1dot = 1
    a2 = 2
    a2dot = 3
    a3 = 4
    a3dot = 5
    θ = 6
    θdot = 7
    ϕ = 8
    ϕdot = 9

    # Dimensionless unit
    x = 0
    xdot = 1
    y = 2
    ydot = 3
    z = 4
    zdot = 5
    # Theta, phi are the same as above.


class InitConIndex(IntEnum):
    """
    Enumerated number for array index for variables in the initial conditions object.
    """
    a1 = 0
    a1dot = 1
    a2 = 2
    a2dot = 3
    a3 = 4
    a3dot = 5
    θ = 6
    θdot = 7
    ϕ = 8
    ϕdot = 9


class EllipIndex(IntEnum):
    """
    Enumerated number for array index for variables in the elliptical integrals
    """
    x = 0
    y = 1
    z = 2
    a1 = 0
    a2 = 1
    a3 = 2


class LengthIndex(IntEnum):
    """
    Enumerator number for array index for a_i lengths of the ellipsoid
    """
    a1 = 0
    a2 = 1
    a3 = 2


def get_BE_equilibrium_radius(mass_r):
    """
    Obtain the two radius solutions to the equilibrium configuration for a BE mass
    :param mass_r, int, goes between 0 and 5 sqrt(5)/16
    used to be Get_BE_Vals[1][0]
    """

    def _eq_rad_dimless(x):
        """
        Calculates the radius for equilibrium.
        RHS of equation where if LHS = 0 then cloud is in equilibirum
        Citation: B.Draine, Physics of the interstellar and intergalactic medium (Princeton University Press,
                  Oxfordshire, 2011
        """
        return x**4 - x * mass_r + 3/5 * mass_r**2

    if mass_r > (5 * sqrt(5))/16:
        print("Selected to use a mass ratio greater than the stable BE mass, returning the local minimum value")
        val_range = linspace(0, 1.5, 2000000)
        val_min, val_idx = min((val, idx) for (idx, val) in enumerate([_eq_rad_dimless(i) for i in val_range]))
        return val_min, val_min

    val_range = linspace(0, 1, 1000000)
    val_min, val_idx = min((val, idx) for (idx, val) in enumerate([_eq_rad_dimless(i) for i in val_range]))
    mp.dps = 100  # Decimal points to use with mp math.

    try:
        # rad_low = findroot(_eq_rad_dimless, val_range[0], solver="muller")
        rad_low = brenth(_eq_rad_dimless, 0, val_range[val_idx] - 1e-10)
    except ValueError:
        rad_low = findroot(_eq_rad_dimless, val_range[0], solver="muller")
        print("ONLY ONE ROOT EXISTS!")
        return float(rad_low), float(rad_low)
    try:
        rad_hig = brenth(_eq_rad_dimless, val_range[val_idx] + 1e-10, val_range[-1])
        print("and at this pressure equilibrium radii of r = {} and {} exist.".format(round(rad_low, 10),
                                                                                      round(rad_hig, 10)),
              "\n" + "The root-finding algorithm can temperamental for the larger root, so check they are different")
    except ValueError:
        print("Root finding algorithm could not find a root to the radius equation."
              "Did you give a proper initial guess?, try changing initial guess, else no equilibrium might exist")
        print("You might want to check which root is failing. The root-finding algorithm is temperamental for the "
              "larger root. Try using more decimal points.")
        rad_low = -1
        rad_hig = -1

    del val_range
    return rad_low, rad_hig


def get_BE_mass_1to4_ratio(percent_be_mass):
    """
    :param percent_be_mass: int. between 0 and 1. The ratio for how massive the cloud should be
    :return int. value between 1-4 representing the density of the Bonnor-Ebert Sphere with respect to the collapse
    density of 4 and the zero cloud mass of 0.
    """
    def eq(x):
        val = percent_be_mass - 1/(5 * sqrt(5)/16) * x * sqrt((x-1)/(3/5 * x**2))**3
        return val

    if 0 < percent_be_mass <= 1:
        mass_r = brenth(eq, 1, 4)

    elif percent_be_mass > 1:
        print("You've selected an unstable cloud that would collapse, Manually set ρ_real_over_ρ_pressure")
        mass_r = -1

    elif percent_be_mass < 0:
        print("Negative mass ratio selected")
        mass_r = -1

    else:
        print("Failed to calculate cloud mass, did you input a interger between 0 and 1? I got {} with type {}".format(percent_be_mass, type(percent_be_mass)))
        mass_r = -1

    return mass_r


def get_BE_mass_0to5sqrt5o16(ρ_normalised, override_percentage=-1):
    """
    Calculate the value between 0 and 5 sqrt(5)/16 for density
    Requires normalised density and returns the mass_ratio required. THe 5sqrt(5)/16 is NOT taken into account
    :param ρ_normalised: int. density of the cloud between 1 and 4 as calculated by get_BE_mass_1to4_ratio
    """
    # todo: far future, fix the printed states so it says 0 to 100% mBE
    if ρ_normalised < 1:
        print("Rho < 1 will result in zero mass specify override percentage")
        return override_percentage

    elif ρ_normalised > 4:
        print('Selected a mass over the maximum mass. Returning maximum value of 5 * sqrt(5)/16. Manually insert value')
        return 5 * sqrt(5)/16

    elif ρ_normalised == 4:
        print("Maximum stable mass selected")
        return 5 * sqrt(5)/16

    else:
        print("Solving for a mass cloud of " +
              str(ρ_normalised * sqrt((ρ_normalised - 1)/(3/5 * ρ_normalised**2))**3) +
              " m_BE, this value is highly non-linear and goes between 0 and approx 0.69877")
        return ρ_normalised * sqrt((ρ_normalised - 1)/(3/5 * ρ_normalised**2))**3


def axis_length_ratio_solver(start, stop):
    """
    Generate the length_2/length_1 axis ratio based on the length_3/length_1 axis ratio which is fed in.
    I.E generate the x/y axis ratio based on the z/x value
    :param start: int, the ratio value which we start solving at for a3/a1
    :param stop: int, the ratio value which we stop solving at for a3/a1
    :return: list of lists: the axis ratio lengths and Chandrasekhar values. Scaled as a_i/R^{(1/3)}
    """
    from scipy.optimize import brenth

    def equ_eq(length1, length2):
        """
        :param length1: int, the value that will be solved for a2/a1
        :param length2: int, the known value, a3/a1
        :return: int, the ratio length_2/length_1 consistent with length_3/length_1
        """
        Ax = get_Ax(x=1, y=length1, z=length2)
        Ay = get_Ay(x=1, y=length1, z=length2)
        Az = get_Az(x=1, y=length1, z=length2)
        return (Ax - Az * length2**2)/(Ay * length1**2 - Az * length2**2) - 3/length2**2 - 1

    vals = linspace(start, stop, 1000)  # This is the range of a3/a1 values we will feed to the functions
    sols = []  # This is a2/a1 values when using a given a3/a1

    for val in vals:
        # I found brenth to be the fastest solver.
        sols.append(brenth(equ_eq, val + 0.0001, 0.9999, args=(val,)))

    plot_vals_a1 = [1/(sols[i] * vals[i])**(1/3) for i in range(len(sols))]   # This gives a1/(a1 a2 a3)^3
    plot_vals_a2 = [1/(vals[i]/sols[i]**2)**(1/3) for i in range(len(sols))]  # This gives a2/(a1 a2 a3)^3
    plot_vals_a3 = [1/(sols[i]/vals[i]**2)**(1/3) for i in range(len(sols))]  # This gives a3/(a1 a2 a3)^3

    # The following values are for p=0 in Chandrasekhar's work.
    chan_a3a1 = [0.91355, 0.80902, 0.66913, 0.54464, 0.50000, 0.48481, 0.46947, 0.45399, 0.40674, 0.32557, 0.30902,
                 0.25882, 0.19081]
    # chan_a2a1 = [0.93188, 0.84112, 0.70687, 0.57787, 0.53013, 0.51373, 0.49714, 0.48040, 0.42898, 0.34052, 0.32254,
    #              0.26827, 0.19569]

    chan_a1 = [1.0551, 1.1369, 1.2835, 1.4701, 1.5567, 1.5894, 1.6242, 1.6613, 1.7896, 2.0816, 2.1568, 2.4330, 2.9919]
    chan_a2 = [0.9832, 0.9563, 0.9072, 0.8495, 0.8253, 0.8165, 0.8074, 0.7981, 0.7677, 0.7088, 0.6957, 0.6527, 0.5855]
    chan_a3 = [0.9639, 0.9198, 0.8588, 0.8007, 0.7784, 0.7706, 0.7625, 0.7542, 0.7279, 0.6777, 0.6665, 0.6297, 0.5709]

    return vals, sols, [plot_vals_a1, plot_vals_a2, plot_vals_a3], [chan_a3a1, chan_a1, chan_a2, chan_a3]


def internal_streaming_axis_length_ratio_solver(alpha):
    """
    Generates the a2/a1 and a3/a1 axis length ratios based on the input internal streaming rate
    Used to be called axis_length_ratio_solver_new from scratch2
    :param alpha: int: The internal streaming value
    :return: ints, a2/a1 and a3/a1
    """

    def a2oa1_eq(alpha):

        return sqrt(1 + 3/(alpha**2))

    def a3oa1_eq(alpha):

        return sqrt(2 * sqrt(alpha**2 + 3) - (alpha**2 + 3))

    return a2oa1_eq(alpha), a3oa1_eq(alpha)


def get_ai_lengths(alpha, ρ, ρ_tidal, return_only_a1=False):
    """
    Compute the dimenless lengths of the a1, a2, and a3 axis, Additionally finds the mass of the cloud
    rho_pressure is normalised out in these equations. The pressure is set to 1
    used to be called get_a3oa1_constant_line_new in scratch3
    :param alpha: int: the internal streaming value
    :param ρ: int: specified density between 1 and 4
    :param ρ_tidal: int: specified tidal strength
    :return: list, individual axis lengths and mass
    """

    def _calcuate_a1_length(ρ, ρ_tidal, a2oa1, a3oa1):
        """
        Calculate the length of the a1 axis using the brent hyperbolic method.
        used to be called get_a1_length_new from scratch3.
        :param ρ: int: density of the cloud, between 1 and 4
        :param ρ_tidal: int: specified tidal strength
        :return:
        """

        def _get_a1_length(a1):
            """
            Equation for root solver
            """
            return a1**2 * a3oa1**2 * ρ_tidal - 5 * (1 - 1 / ρ)

        return brenth(_get_a1_length, 0.001, 300)

    # Obtain the a2/a1 and a3/a1 axis ratios for a given alpha.
    a2_over_a1, a3_over_a1 = internal_streaming_axis_length_ratio_solver(alpha)

    # Calculate the ai lengths
    a1_len = _calcuate_a1_length(ρ, ρ_tidal, a2_over_a1, a3_over_a1)
    a2_len = a2_over_a1 * a1_len
    a3_len = a3_over_a1 * a1_len
    # and mass
    mass = 4/3 * ρ * a1_len * a2_len * a3_len

    if return_only_a1:
        return a1_len
    else:
        return [a1_len, a2_len, a3_len], mass


def get_ai_lengths_chandrasekhar(a3_over_a1_index, specific_ρ, a3_over_a1_list=False, a2_over_a1_list=False):
    """
    Generate the ai lengths for the cloud based on the work by Chandrasekhar. Additionally calculates mass, mu, and
    tidal strength
    :param a3_over_a1_index. int. This is a tricky param. other methods need to be employed to find the correct index to
    give the correct a3/a1 length which is calculated by the 1000 index list generated by the function
    axis_length_ratio_solver - best to consult the author.
    :param specific_ρ: Choose a specific rho between 1 and 4 to use
    :return:
    """

    def _calculate_a1_length_chandrasekhar(ρ, ρ_tidal, a2_over_a1, a3_over_a1):
        """
        Calculate the length of the a1 axis as specified by Chandrasekhar
        :param ρ: int: density
        :param ρ_tidal: int: tidal strength
        :param a2_over_a1: int: a2 over a1 axis ratio
        :param a3_over_a1: int: a3 over a1 axis ratio
        :return: the a1 length as found by mullers method.
        """

        def _get_a1_length(a1):
            """
            Equation for the root solver
            """
            return a1 ** 2 * (3 * ρ_tidal / 1 - 9 / 2 * ρ / 1 * a2_over_a1 * a3_over_a1 * Ax) + 5 * (1 - 1 / ρ)

        Ax = get_Ax(x=1, y=a2_over_a1, z=a3_over_a1)

        return findroot(_get_a1_length, 0.5, solver="muller")  # Can instead do return brenth(_a1_length_equ, 0, 3)

    def _get_Omega_Chan(a2_over_a1, a3_over_a1):
        """
        Calculates the value for mu as specified in Chandrasekhar (year)
        :param a2oa1:
        :param a3oa1:
        :return:
        """
        return 2 * a2_over_a1 * a3_over_a1 * (get_Ax(x=1, y=a2_over_a1, z=a3_over_a1) - get_Az(x=1, y=a2_over_a1, z=a3_over_a1) * a3_over_a1 ** 2) / (
                    3 + a3_over_a1 ** 2)

    # generate the ratios and select the desired value
    if not a3_over_a1_list and not a2_over_a1_list:
        print("No a3/a1 and/or a2/a1 list provided - Calculating")
        a3_over_a1_list, a2_over_a1_list, _, _ = axis_length_ratio_solver(0.04, 0.99)
    a3_over_a1 = a3_over_a1_list[a3_over_a1_index]
    a2_over_a1 = a2_over_a1_list[a3_over_a1_index]

    # Calculate mu and rho tides
    mu_val = _get_Omega_Chan(a2_over_a1, a3_over_a1)  # in units of pi G rho. Explicitly it is mu/(pi G rho)
    ρ_tidal = (9/(4 * pi) * (pi * specific_ρ) * mu_val)

    # Calculate the axis lengths
    a1_len = _calculate_a1_length_chandrasekhar(specific_ρ, ρ_tidal, a2_over_a1, a3_over_a1)
    a2_len = a2_over_a1 * a1_len
    a3_len = a3_over_a1 * a1_len

    # and the mass
    mass_val = 4/3 * specific_ρ * a2_over_a1 * a3_over_a1 * a1_len**3

    return [float(a1_len), float(a2_len), float(a3_len)], float(mass_val), [specific_ρ, ρ_tidal, mu_val],


def get_physical_scale(solution, times, init_con, temp, pext, l_unit="cm"):
    """
    Calculate the physical scale of the cloud based on an input temperature and external pressure
    :param solution: Solution array
    :param times: list; list of times corresponding to the solver points
    :param init_con: InitialConditions object
    :param temp: int. Temperature in Kelvin
    :param pext: int: External pressure in Kelvin per cubic centimetre
    :param l_unit: string: the length unit to use, either pc (parsec), m (metre), or cm (centimetre) if defaulted.
    :return: SolutionPhysicalUnits: solution class object with physical parameters.
    """

    from src.solve.utils import get_cs2
    from src.physics_constants import GRAV, BOLTZMANN
    from src.data_formats import SolutionPhysicalUnits

    pext = pext * BOLTZMANN  # Convert from Kelvin per cubic centimetre to pascals
    cs2 = get_cs2(temp=temp, mu=2.33)  # Returns in units of cm^2 s^-2

    len_scale = 3 * cs2 * 1/sqrt(4 * pi * GRAV * pext)       # Returns in units of cm
    time_scale = 3 * sqrt(cs2/(4 * pi * GRAV * pext))        # Returns in units of seconds
    mass_scale = 9 * cs2**2/(sqrt(4 * pi * GRAV**3 * pext))  # Returns in units of grams

    if l_unit.lower() == "pc":
        len_scale = len_scale * 3.240779289e-19
    if l_unit.lower() == "m":
        len_scale = len_scale * 0.01

    # give everything its corresponding time scale
    for index in (ODEIndex.xdot, ODEIndex.ydot, ODEIndex.zdot, ODEIndex.θdot, ODEIndex.ϕdot):
        solution[:, index] = solution[:, index] / time_scale
        times[:] = times[:] / time_scale
    # and length unit
    for index in (ODEIndex.x, ODEIndex.xdot, ODEIndex.y, ODEIndex.ydot, ODEIndex.z, ODEIndex.zdot):
        solution[:, index] = solution[:, index] * len_scale

    # todo: pull the densitiy and mass from init con and such.

    mass = init_con.mass_r * 1


    # # Calculate the respective densities
    # soln = SolutionPhysicalUnits(
    #     times=times,
    #     solution=solution,
    #     ρ_real=)




def calculate_solution_planes(x_val, y_val, z_val, θ, θdot, ϕdot, ρ_self, ρ_tides, style_heat=True, log_scale=True, constant_var="x"):
    """ calculate_solution_planes(0.64838746, 0.36108903, 0.34756538, 0, 0, -0.1/sqrt(ρ_pressure_over_ρ_tides), 1.1181313072462846, 1/4.4, "x")
    Calculate and plot the solution plane for a constant value and then ranging two other
    :return:
    """
    import matplotlib.pyplot as plt
    import numpy as np
    from src.solve.deriv_funcs import deriv_xdot_func, deriv_ydot_func, deriv_zdot_func
    from src.solve.geometry import get_Ax, get_Ay, get_Az
    from matplotlib import cm
    from mpl_toolkits.mplot3d import Axes3D
    import cmasher as cmr
    cmap = cmr.iceburn
    nums = 100
    up_down_val = 0.03
    vmin_max_val = 0.03
    x_range = np.linspace(x_val - up_down_val, x_val + up_down_val, nums)
    y_range = np.linspace(y_val - up_down_val, y_val + up_down_val, nums)
    z_range = np.linspace(z_val - up_down_val, z_val + up_down_val, nums)

    if constant_var.lower() == "x":
        xddot_vals = []
        yddot_vals = []
        zddot_vals = []
        sumofsqs = []
        for y in y_range:
            for z in z_range:
                Ai_vals = [get_Ax(x=x_val, y=y, z=z), get_Ay(x=x_val, y=y, z=z), get_Az(x=x_val, y=y, z=z)]
                xddot_vals.append(deriv_xdot_func(
                    x=x_val, y=y, z=z, θ=θ, θdot=θdot, ϕdot=ϕdot, A1=Ai_vals[EllipIndex.x],
                    ρ_real_over_ρ_pressure=ρ_self, ρ_pressure_over_ρ_tides=1/ρ_tides)[0])
                yddot_vals.append(deriv_ydot_func(
                    x=x_val, y=y, z=z, θ=θ, θdot=θdot, ϕdot=ϕdot, A2=Ai_vals[EllipIndex.y],
                    ρ_real_over_ρ_pressure=ρ_self, ρ_pressure_over_ρ_tides=1/ρ_tides)[0])
                zddot_vals.append(deriv_zdot_func(
                    x=x_val, y=y, z=z, A3=Ai_vals[EllipIndex.z],
                    ρ_real_over_ρ_pressure=ρ_self, ρ_pressure_over_ρ_tides=1/ρ_tides)[0])
        for (x, y, z) in zip(xddot_vals, yddot_vals, zddot_vals):
            sumofsqs.append(x**2 + y**2 + z**2)

        plot_xddot = np.flipud(np.reshape(xddot_vals, (nums, nums)).T)
        plot_yddot = np.flipud(np.reshape(yddot_vals, (nums, nums)).T)
        plot_zddot = np.flipud(np.reshape(zddot_vals, (nums, nums)).T)
        plot_sumofsqs = np.flipud(np.reshape(sumofsqs, (nums, nums)).T)

        if log_scale:
            plot_xddot = np.log10(np.abs(plot_xddot))
            plot_yddot = np.log10(np.abs(plot_yddot))
            plot_zddot = np.log10(np.abs(plot_zddot))
            plot_sumofsqs = np.log10(plot_sumofsqs)

        if style_heat:

            min_plot = np.min(np.concatenate((plot_xddot, plot_yddot, plot_zddot)))
            max_plot = np.max(np.concatenate((plot_xddot, plot_yddot, plot_zddot)))

            fig_heatmaps, ((ax1, ax2, ax3), (ax4, ax5, ax6)) = plt.subplots(2, 3, sharey=True, sharex=True)
            f1 = ax1.imshow(plot_xddot, extent=(y_range[0], y_range[-1], z_range[0], z_range[-1]), vmin=min_plot, vmax=max_plot, cmap=cmap)
            f2 = ax2.imshow(plot_yddot, extent=(y_range[0], y_range[-1], z_range[0], z_range[-1]) , vmin=min_plot, vmax=max_plot, cmap=cmap)
            f3 = ax3.imshow(plot_zddot, extent=(y_range[0], y_range[-1], z_range[0], z_range[-1]) , vmin=min_plot, vmax=max_plot, cmap=cmap)
            f4 = ax4.imshow(plot_sumofsqs, extent=(y_range[0], y_range[-1], z_range[0], z_range[-1]) , vmin=np.min(plot_sumofsqs), vmax=np.max(plot_sumofsqs))
            fig_heatmaps.suptitle("Constant x = {}, alpha = {}".format(x_val, ϕdot*sqrt(1/ρ_tides)))
            ax1.set_xlabel("y value")
            ax2.set_xlabel("y value")
            ax3.set_xlabel("y value")
            ax4.set_xlabel("y value")
            ax1.set_ylabel("z value")
            ax4.set_ylabel("z value")
            ax1.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=False, direction="inout")
            ax2.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=False, direction="inout")
            ax3.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=True,  direction="inout")
            ax4.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=True, direction="inout")
            ax1.plot(y_val, z_val, color='white', marker="+", ms=12)
            ax2.plot(y_val, z_val, color='white', marker="+", ms=12)
            ax3.plot(y_val, z_val, color='white', marker="+", ms=12)
            ax4.plot(y_val, z_val, color='white', marker="+", ms=12)
            fig_heatmaps.delaxes(ax5)
            fig_heatmaps.delaxes(ax6)
            fig_heatmaps.subplots_adjust(wspace=0)
            fig_heatmaps.colorbar(f1, label="Log10 code value", aspect=5, cmap=cmap, ax=ax5)
            fig_heatmaps.colorbar(f4, label="Log10 code value", aspect=5, cmap=cmr.voltage, ax=ax6)
            ax1.set_title("xddot")
            ax2.set_title("yddot")
            ax3.set_title("zddot")
            ax4.set_title("sum in quad.")

        else:
        # Plot a 3d grid of the 3 planes

            fig = plt.figure()
            ax = fig.gca(projection='3d')
            plot_y, plot_z = np.meshgrid(y_range, z_range)

            graph_xddot = ax.plot_surface(plot_y, plot_z, plot_xddot, cmap=cm.PuOr, linewidth=0, antialiased=False, vmin=-vmin_max_val, vmax=vmin_max_val)
            graph_yddot = ax.plot_surface(plot_y, plot_z, plot_yddot, cmap=cm.PiYG, linewidth=0, antialiased=False, vmin=-vmin_max_val, vmax=vmin_max_val)
            graph_zddot = ax.plot_surface(plot_y, plot_z, plot_zddot, cmap=cm.bwr,  linewidth=0, antialiased=False, vmin=-vmin_max_val, vmax=vmin_max_val)
            fig.colorbar(graph_xddot, shrink=0.5, aspect=5)
            fig.colorbar(graph_yddot, shrink=0.5, aspect=5)
            fig.colorbar(graph_zddot, shrink=0.5, aspect=5)

            plt.title("Constant x = {}, alpha = {}".format(x_val, ϕdot*sqrt(1/ρ_tides)))
            plt.xlabel("y value")
            plt.ylabel("z value")
            ax.set_zlabel("ddot value")



    elif constant_var.lower() == "y":
        xddot_vals = []
        yddot_vals = []
        zddot_vals = []
        sumofsqs = []

        for x in x_range:
            for z in z_range:
                Ai_vals = [get_Ax(x=x, y=y_val, z=z), get_Ay(x=x, y=y_val, z=z), get_Az(x=x, y=y_val, z=z)]
                xddot_vals.append(deriv_xdot_func(
                    x=x, y=y_val, z=z, θ=θ, θdot=θdot, ϕdot=ϕdot, A1=Ai_vals[EllipIndex.x],
                    ρ_real_over_ρ_pressure=ρ_self, ρ_pressure_over_ρ_tides=1/ρ_tides)[0])
                yddot_vals.append(deriv_ydot_func(
                    x=x, y=y_val, z=z, θ=θ, θdot=θdot, ϕdot=ϕdot, A2=Ai_vals[EllipIndex.y],
                    ρ_real_over_ρ_pressure=ρ_self, ρ_pressure_over_ρ_tides=1/ρ_tides)[0])
                zddot_vals.append(deriv_zdot_func(
                    x=x, y=y_val, z=z, A3=Ai_vals[EllipIndex.z],
                    ρ_real_over_ρ_pressure=ρ_self, ρ_pressure_over_ρ_tides=1/ρ_tides)[0])
        for (x, y, z) in zip(xddot_vals, yddot_vals, zddot_vals):
            sumofsqs.append(x**2 + y**2 + z**2)

        plot_xddot = np.flipud(np.reshape(xddot_vals, (nums, nums)))
        plot_yddot = np.flipud(np.reshape(yddot_vals, (nums, nums)))
        plot_zddot = np.flipud(np.reshape(zddot_vals, (nums, nums)))
        plot_sumofsqs = np.flipud(np.reshape(sumofsqs, (nums, nums)))

        if log_scale:
            plot_xddot = np.log10(np.abs(plot_xddot))
            plot_yddot = np.log10(np.abs(plot_yddot))
            plot_zddot = np.log10(np.abs(plot_zddot))
            plot_sumofsqs = np.log10(plot_sumofsqs)

        if style_heat:

            min_plot = np.min(np.concatenate((plot_xddot, plot_yddot, plot_zddot)))
            max_plot = np.max(np.concatenate((plot_xddot, plot_yddot, plot_zddot)))

            fig_heatmaps, ((ax1, ax2, ax3), (ax4, ax5, ax6)) = plt.subplots(2, 3, sharey=True, sharex=True)
            f1 = ax1.imshow(plot_xddot, extent=(x_range[0], x_range[-1], z_range[0], z_range[-1]), vmin=min_plot, vmax=max_plot, cmap=cmap)
            f2 = ax2.imshow(plot_yddot, extent=(x_range[0], x_range[-1], z_range[0], z_range[-1]) , vmin=min_plot, vmax=max_plot, cmap=cmap)
            f3 = ax3.imshow(plot_zddot, extent=(x_range[0], x_range[-1], z_range[0], z_range[-1]) , vmin=min_plot, vmax=max_plot, cmap=cmap)
            f4 = ax4.imshow(plot_sumofsqs, extent=(x_range[0], x_range[-1], z_range[0], z_range[-1]) , vmin=np.min(plot_sumofsqs), vmax=np.max(plot_sumofsqs))
            fig_heatmaps.suptitle("Constant y = {}, alpha = {}".format(y_val, ϕdot*sqrt(1/ρ_tides)))
            ax1.set_xlabel("x value")
            ax2.set_xlabel("x value")
            ax3.set_xlabel("x value")
            ax4.set_xlabel("x value")
            ax1.set_ylabel("z value")
            ax4.set_ylabel("sum in quad.")
            ax1.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=False, direction="inout")
            ax2.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=False, direction="inout")
            ax3.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=True,  direction="inout")
            ax4.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=True, direction="inout")
            ax1.plot(x_val, z_val, color='white', marker="+", ms=12)
            ax2.plot(x_val, z_val, color='white', marker="+", ms=12)
            ax3.plot(x_val, z_val, color='white', marker="+", ms=12)
            ax4.plot(x_val, z_val, color='white', marker="+", ms=12)
            fig_heatmaps.delaxes(ax5)
            fig_heatmaps.delaxes(ax6)
            fig_heatmaps.subplots_adjust(wspace=0)
            fig_heatmaps.colorbar(f1, label="Log10 code value", aspect=5, cmap=cmap, ax=ax5)
            fig_heatmaps.colorbar(f4, label="Log10 code value", aspect=5, cmap=cmr.voltage, ax=ax6)
            ax1.set_title("xddot")
            ax2.set_title("yddot")
            ax3.set_title("zddot")
            ax4.set_title("sum in quad.")

        else:
            fig = plt.figure()
            ax = fig.gca(projection='3d')
            plot_x, plot_z = np.meshgrid(x_range, z_range)

            graph_xddot = ax.plot_surface(plot_x, plot_z, plot_xddot, cmap=cm.PuOr, linewidth=0, antialiased=False,
                                          vmin=-vmin_max_val, vmax=vmin_max_val)
            graph_yddot = ax.plot_surface(plot_x, plot_z, plot_yddot, cmap=cm.PiYG, linewidth=0, antialiased=False,
                                          vmin=-vmin_max_val, vmax=vmin_max_val)
            graph_zddot = ax.plot_surface(plot_x, plot_z, plot_zddot, cmap=cm.bwr, linewidth=0, antialiased=False,
                                          vmin=-vmin_max_val, vmax=vmin_max_val)
            fig.colorbar(graph_xddot, shrink=0.5, aspect=5)
            fig.colorbar(graph_yddot, shrink=0.5, aspect=5)
            fig.colorbar(graph_zddot, shrink=0.5, aspect=5)

            plt.title("Constant y = {}, alpha = {}".format(y_val, ϕdot*sqrt(1/ρ_tides)))
            plt.xlabel("x value")
            plt.ylabel("z value")
            ax.set_zlabel("ddot value")

    elif constant_var.lower() == "z":
        xddot_vals = []
        yddot_vals = []
        zddot_vals = []
        sumofsqs = []
        for x in x_range:
            for y in y_range:
                Ai_vals = [get_Ax(x=x, y=y, z=z_val), get_Ay(x=x, y=y, z=z_val), get_Az(x=x, y=y, z=z_val)]
                xddot_vals.append(deriv_xdot_func(
                    x=x, y=y, z=z_val, θ=θ, θdot=θdot, ϕdot=ϕdot, A1=Ai_vals[EllipIndex.x],
                    ρ_real_over_ρ_pressure=ρ_self, ρ_pressure_over_ρ_tides=1/ρ_tides)[0])
                yddot_vals.append(deriv_ydot_func(
                    x=x, y=y, z=z_val, θ=θ, θdot=θdot, ϕdot=ϕdot, A2=Ai_vals[EllipIndex.y],
                    ρ_real_over_ρ_pressure=ρ_self, ρ_pressure_over_ρ_tides=1/ρ_tides)[0])
                zddot_vals.append(deriv_zdot_func(
                    x=x, y=y, z=z_val, A3=Ai_vals[EllipIndex.z],
                    ρ_real_over_ρ_pressure=ρ_self, ρ_pressure_over_ρ_tides=1/ρ_tides)[0])
        for (x, y, z) in zip(xddot_vals, yddot_vals, zddot_vals):
            sumofsqs.append(x**2 + y**2 + z**2)

        plot_xddot = np.flipud(np.reshape(xddot_vals, (nums, nums)))
        plot_yddot = np.flipud(np.reshape(yddot_vals, (nums, nums)))
        plot_zddot = np.flipud(np.reshape(zddot_vals, (nums, nums)))
        plot_sumofsqs = np.flipud(np.reshape(sumofsqs, (nums, nums)))

        if log_scale:
            plot_xddot = np.log10(np.abs(plot_xddot))
            plot_yddot = np.log10(np.abs(plot_yddot))
            plot_zddot = np.log10(np.abs(plot_zddot))
            plot_sumofsqs = np.log10(plot_sumofsqs)

        if style_heat:

            min_plot = np.min(np.concatenate((plot_xddot, plot_yddot, plot_zddot)))
            max_plot = np.max(np.concatenate((plot_xddot, plot_yddot, plot_zddot)))

            fig_heatmaps, ((ax1, ax2, ax3), (ax4, ax5, ax6)) = plt.subplots(2, 3, sharey=True, sharex=True)
            f1 = ax1.imshow(plot_xddot, extent=(x_range[0], x_range[-1], y_range[0], y_range[-1]), vmin=min_plot, vmax=max_plot, cmap=cmap)
            f2 = ax2.imshow(plot_yddot, extent=(x_range[0], x_range[-1], y_range[0], y_range[-1]) , vmin=min_plot, vmax=max_plot, cmap=cmap)
            f3 = ax3.imshow(plot_zddot, extent=(x_range[0], x_range[-1], y_range[0], y_range[-1]) , vmin=min_plot, vmax=max_plot, cmap=cmap)
            f4 = ax4.imshow(plot_sumofsqs, extent=(x_range[0], x_range[-1], y_range[0], y_range[-1]) , vmin=np.min(plot_sumofsqs), vmax=np.max(plot_sumofsqs))
            fig_heatmaps.suptitle("Constant z = {}, alpha = {}".format(z_val, ϕdot*sqrt(1/ρ_tides)))
            ax1.set_xlabel("x value")
            ax2.set_xlabel("x value")
            ax3.set_xlabel("x value")
            ax4.set_xlabel("x value")
            ax1.set_ylabel("y value")
            ax4.set_ylabel("sum in quad.")
            ax1.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=False, direction="inout")
            ax2.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=False, direction="inout")
            ax3.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=True,  direction="inout")
            ax4.tick_params(top=True, right=True, labelbottom=True, left=True, labelright=True, direction="inout")
            ax1.plot(x_val, y_val, color='white', marker="+", ms=12)
            ax2.plot(x_val, y_val, color='white', marker="+", ms=12)
            ax3.plot(x_val, y_val, color='white', marker="+", ms=12)
            ax4.plot(x_val, y_val, color='white', marker="+", ms=12)
            fig_heatmaps.delaxes(ax5)
            fig_heatmaps.delaxes(ax6)
            fig_heatmaps.subplots_adjust(wspace=0)
            fig_heatmaps.colorbar(f1, label="Log10 code value", aspect=5, cmap=cmap, ax=ax5)
            fig_heatmaps.colorbar(f4, label="Log10 code value", aspect=5, cmap=cmr.voltage, ax=ax6)
            ax1.set_title("xddot")
            ax2.set_title("yddot")
            ax3.set_title("zddot")
            ax4.set_title("sum in quad.")

        else:
            fig = plt.figure()
            ax = fig.gca(projection='3d')
            plot_x, plot_y = np.meshgrid(x_range, y_range)

            graph_xddot = ax.plot_surface(plot_x, plot_y, plot_xddot, cmap=cm.PuOr, linewidth=0, antialiased=False,
                                          vmin=-vmin_max_val, vmax=vmin_max_val)
            graph_yddot = ax.plot_surface(plot_x, plot_y, plot_yddot, cmap=cm.PiYG, linewidth=0, antialiased=False,
                                          vmin=-vmin_max_val, vmax=vmin_max_val)
            graph_zddot = ax.plot_surface(plot_x, plot_y, plot_zddot, cmap=cm.bwr, linewidth=0, antialiased=False,
                                          vmin=-vmin_max_val, vmax=vmin_max_val)
            fig.colorbar(graph_xddot, shrink=0.5, aspect=5)
            fig.colorbar(graph_yddot, shrink=0.5, aspect=5)
            fig.colorbar(graph_zddot, shrink=0.5, aspect=5)

            plt.title("Constant z = {}, alpha = {}".format(z_val, ϕdot*sqrt(1/ρ_tides)))
            plt.xlabel("x value")
            plt.ylabel("y value")
            ax.set_zlabel("ddot value")


    else:
        raise SystemExit("Incorrect variable selected, got {}, expected x, y, or z".format(constant_var))

    # fig, ax = plt.subplot(1, 3)
    # calculate_solution_planes(0.5, 0.5, 0.5, 0, 0, -0.1, 1.18, 0.01, "x")





"""
Functions related to obtaining the equilibrium starting conditions for ellipsoidal clouds with internal streaming.
"""
import numpy as np
import scipy.optimize as sco
from math import sqrt
from enum import IntEnum
import mpmath as mp

from src.solve.geometry import get_Ax, get_Az, get_Ay
from src.utils import EllipIndex, internal_streaming_axis_length_ratio_solver, get_ai_lengths
from src.solve.deriv_funcs import deriv_xdot_func, deriv_ydot_func, deriv_zdot_func


class VarIndex(IntEnum):
    """
    Enum to clean up vectorised derivative equations
    """
    x = 0
    y = 1
    z = 2


# Derivative equations for the ellipsoid in in vector form
def deriv_xdot_func_vec(vec_input, θ, θdot, ϕdot, A1, ρ_real_over_ρ_pressure, ρ_pressure_over_ρ_tides):
    """
    Compute the derivative of the derivative of the axis of the ellipsoid
    """
    return deriv_xdot_func(x=vec_input[VarIndex.x], y=vec_input[VarIndex.y], z=vec_input[VarIndex.z], A1=A1, θ=θ, θdot=θdot,
                           ϕdot=ϕdot, ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure,
                           ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)[0]


def deriv_ydot_func_vec(vec_input, θ, θdot, ϕdot, A2, ρ_real_over_ρ_pressure, ρ_pressure_over_ρ_tides):
    """
    Compute the derivative of the derivative of the axis of the ellipsoid ellipsoid
    """
    return deriv_ydot_func(x=vec_input[VarIndex.x], y=vec_input[VarIndex.y], z=vec_input[VarIndex.z], A2=A2, θ=θ, θdot=θdot,
                           ϕdot=ϕdot, ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure,
                           ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)[0]


def deriv_zdot_func_vec(vec_input, A3, ρ_real_over_ρ_pressure, ρ_pressure_over_ρ_tides):
    """
    Compute the derivative of the derivative of the axis of the ellipsoid
    """
    return deriv_zdot_func(x=vec_input[VarIndex.x], y=vec_input[VarIndex.y], z=vec_input[VarIndex.z], A3=A3,
                           ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure,
                           ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)[0]


def combined_deriv_eqs(x, ρ_real_over_ρ_pressure, ρ_pressure_over_ρ_tides, ϕdot):
    """
    List of vector form of derivative equations for scipy root
    """
    Ai = [
        get_Ax(x=x[VarIndex.x], y=x[VarIndex.y], z=x[VarIndex.z]),
        get_Ay(x=x[VarIndex.x], y=x[VarIndex.y], z=x[VarIndex.z]),
        get_Az(x=x[VarIndex.x], y=x[VarIndex.y], z=x[VarIndex.z])]

    return [
        deriv_xdot_func_vec(vec_input=x, θ=0, θdot=0, ϕdot=ϕdot, A1=Ai[EllipIndex.x],
                            ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure,
                            ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides),
        deriv_ydot_func_vec(vec_input=x, θ=0, θdot=0, ϕdot=ϕdot, A2=Ai[EllipIndex.y],
                            ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure,
                            ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides),
        deriv_zdot_func_vec(vec_input=x, A3=Ai[EllipIndex.z],
                            ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure,
                            ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)
    ]


def combined_deriv_eqs_quad(x, ρ_real_over_ρ_pressure, ρ_pressure_over_ρ_tides, ϕdot):
    """
    Vector form of derivative equations for scipy minimise to minimise the equations in quadrature.
    """
    Ai = [
        get_Ax(x=x[VarIndex.x], y=x[VarIndex.y], z=x[VarIndex.z]),
        get_Ay(x=x[VarIndex.x], y=x[VarIndex.y], z=x[VarIndex.z]),
        get_Az(x=x[VarIndex.x], y=x[VarIndex.y], z=x[VarIndex.z])]
    return deriv_xdot_func_vec(vec_input=x, θ=0, θdot=0, ϕdot=ϕdot, A1=Ai[EllipIndex.x],
                                ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure,
                                ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)**2 + \
            deriv_ydot_func_vec(vec_input=x, θ=0, θdot=0, ϕdot=ϕdot, A2=Ai[EllipIndex.y],
                                ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure,
                                ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)**2 + \
            deriv_zdot_func_vec(vec_input=x, A3=Ai[EllipIndex.z],
                                ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure,
                                ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)**2


def get_rot_equ_axis_lengths(alpha, rho_lim, ρ_tides, ρ_init=1.0001, small_equ_set=True, final_only=True):
    """
    Calculates the equilibrium axis lengths based on the full ODE system for a ellipsoid with a given density, tidal
    density and with an internal streaming velocity. Initial conditions are initially calculated with no gravity.
    Following this the equilibrium solution to solved by "hopping" to the desired ellipsoidal conditions through solving
    the full ODE system with small increments in its parameters.
    :param alpha: int: internal streaming rate
    :param rho_lim: int: the maximum density value to be calculated. This should be between 1 and 4.
    :param ρ_tides: int: The tidal density
    :param ρ_init: int: The initial density of the cloud - does not need to be changed
    :param small_equ_set: bool: determine to use the larger or the smaller (unstable/stable) equ solutions.
    :param final_only: bool: return only the final (rho_lim) values for axis lengths
    :return:
    """

    if alpha == 0:
        raise SystemExit("Alpha was zero, select a non-zero value and rerun. Exiting. ")

    # Constants
    ϕdot = alpha * sqrt(ρ_tides)
    # Get the axis ratios and the lengths for the first step in which no gravity is considered to be acting.
    a2oa1, a3oa1 = internal_streaming_axis_length_ratio_solver(alpha)
    a1 = get_ai_lengths(alpha=alpha, ρ=ρ_init, ρ_tidal=ρ_tides, return_only_a1=True)
    a2 = a2oa1 * a1
    a3 = a3oa1 * a1
    no_grav_a1 = a1

    # Prepare the parameters to solve the ODE system.
    indexs_calculated = 0
    ai_list = []
    ρ_list = []
    flag_list = []
    # and solve for the first actual length of a1 WHERE GRAVITY IS INCLUDED.
    a1_new = sco.root(fun=combined_deriv_eqs, x0=np.array([a1, a2, a3]),
                      args=(ρ_init, 1/ρ_tides, ϕdot), method='anderson',
                      options={"maxiter": 1000, "fatol": 1e-13})
    # append the values to lists
    ai_list.append(a1_new.x)
    ρ_list.append(ρ_init)
    flag_list.append(a1_new.success)
    indexs_calculated += 1

    # Begin solving the system to the desired parameters.
    while True:
        a1_new = sco.root(fun=combined_deriv_eqs, x0=a1_new.x, args=(ρ_init, 1/ρ_tides, ϕdot),
                          method='lm',
                          options={"xtol": 1e-12, "maxfev": 100000, "ftol": 1e-5, "maxiter": 25000})
        # a1_new = sco.minimize(fun=combined_deriv_eqs_quad, x0=a1_new.x, args=(ρ_init, 1/ρ_tides, ϕdot), tol=1e-12, method="Nelder-Mead", options={"adaptive":True, "xatol": 1e-12, "maxfev": 100000, "fatol":1e-12})
        # a1_new = sco.differential_evolution(func=combined_deriv_eqs_quad, bounds=[(0.2, 1), (0.2, 1), (0.2, 1)], args=(ρ_init, 1/ρ_tides, ϕdot), tol=1e-12)
        indexs_calculated += 1
        ai_list.append(a1_new.x)
        ρ_list.append(ρ_init)
        flag_list.append(a1_new.success)

        if ρ_init > rho_lim:
            print("All values to limit were calculated")
            break

        # required for large equ set.
        prev_a1 = ai_list[-1][VarIndex.x]
        prev_a2 = ai_list[-1][VarIndex.y]
        prev_a3 = ai_list[-1][VarIndex.z]

        if small_equ_set:
            # If the solver found a negative root we invert it for the next step. NOTE. The negative values are saved
            # for plotting. They are not discarded and the absolute value taken.
            if a1_new.x[VarIndex.x] < 0:
                a1_new.x[VarIndex.x] = - a1_new.x[VarIndex.x]
            if a1_new.x[VarIndex.y] < 0:
                a1_new.x[VarIndex.y] = - a1_new.x[VarIndex.y]
            if a1_new.x[VarIndex.z] < 0:
                a1_new.x[VarIndex.z] = - a1_new.x[VarIndex.z]

            # Currently we look for solutions with lengths less than 5 - This is tides 0.001. Hence we reset the value
            if a1_new.x[VarIndex.x] > 5:
                a1_new.x[VarIndex.x] = 0.7
            if a1_new.x[VarIndex.y] > 1:
                a1_new.x[VarIndex.y] = 0.2
            if a1_new.x[VarIndex.z] > 5:
                a1_new.x[VarIndex.z] = 0.7

        elif not small_equ_set:
            # In this case if the axis lengths jump to negative values assign them the previous value.
            if a1_new.x[VarIndex.x] <= 0:
                a1_new.x[VarIndex.x] = prev_a1
            if a1_new.x[VarIndex.y] <= 0:
                a1_new.x[VarIndex.y] = prev_a2
            if a1_new.x[VarIndex.z] <= 0:
                a1_new.x[VarIndex.z] = prev_a3
            # In the case that the solver didn't find a new value we will increase the length by a bit, read 1 unit,
            # for the next guess.
            if prev_a1 == a1_new.x[VarIndex.x]:
                a1_new.x[VarIndex.x] = prev_a1 + 1
            if prev_a2 == a1_new.x[VarIndex.y]:
                a1_new.x[VarIndex.y] = prev_a1 + 1
            if prev_a3 == a1_new.x[VarIndex.z]:
                a1_new.x[VarIndex.z] = prev_a1 + 1

        else:
            print("Did not select which equilibrium set to use. exiting")
            raise SystemExit

        # increment density and back through the loop we go
        ρ_init = ρ_init + 1/2000

    if final_only:
        ai_list = ai_list[-1]
        ρ_list = ρ_list[-1]
        mass = 4 / 3 * ρ_list * ai_list[0] * ai_list[1] * ai_list[2]
    else:
        mass = [4 / 3 * ρ * a1 * a2 * a3 for ρ, (a1, a2, a3) in zip(ρ_list, ai_list)]

    return ai_list, mass, [ρ_list, indexs_calculated, no_grav_a1, flag_list]


def index_finder(ρ_list, ρ_wanted):
    """Returns the closest density without going over"""
    for index, ρ in enumerate(ρ_list):
        if ρ < ρ_wanted:
            try:
                if ρ_list[index + 1] > ρ_wanted:
                    return index
            except IndexError:
                print("next rho val is outside of list")
                return -1



"""
Define the data structures used.
"""
import attr
from numpy import asarray, concatenate

from src.utils import ODEIndex


@attr.s
class InitialConditions:
    """
    Datastructure for the initial conditions of the system and ODE
    """
    ρ_real_over_ρ_tides = attr.ib()
    ρ_pressure_over_ρ_tides = attr.ib()
    ρ_real_over_ρ_pressure = attr.ib()
    mass_r = attr.ib()
    equ_radius = attr.ib()
    ode_init_con = attr.ib(default=attr.Factory(list))
    after_tstop_params = attr.ib(default=attr.Factory(list))
    @ode_init_con.validator
    def check(self, attribute, values):
        if len(values) != 10:
            raise RuntimeError("ODE requires 10 initial conditions, {} were given".format(len(self.ode_init_con)))


@attr.s
class ConfigParams:
    """
    Datastructure for the runtime configuration options.
    """
    start = attr.ib()
    stop = attr.ib()
    num_time = attr.ib()
    max_steps = attr.ib()
    relative_tolerance = attr.ib()
    absolute_tolerance = attr.ib()
    enable_taylor_jump = attr.ib()
    taylor_jumps_num = attr.ib()
    taylor_jump_threshold = attr.ib()
    enable_tstop = attr.ib()
    tstop_times = attr.ib(default=attr.Factory(list))


@attr.s
class Solution:
    """
    Datastructure of solutions.
    """
    times = attr.ib()
    solution = attr.ib()
    flags = attr.ib(default=attr.Factory(list))
    tstop_times = attr.ib(default=attr.Factory(list))
    jump_times = attr.ib(default=attr.Factory(list))


@attr.s
class InternalData:
    """
    Datastructure for internally solved params
    """
    params = attr.ib(default=attr.Factory(list))
    derivs = attr.ib(default=attr.Factory(list))
    times = attr.ib(default=attr.Factory(list))
    a1 = attr.ib(default=attr.Factory(list))
    a1dot = attr.ib(default=attr.Factory(list))
    a2 = attr.ib(default=attr.Factory(list))
    a2dot = attr.ib(default=attr.Factory(list))
    a3 = attr.ib(default=attr.Factory(list))
    a3dot = attr.ib(default=attr.Factory(list))
    θ = attr.ib(default=attr.Factory(list))
    θdot = attr.ib(default=attr.Factory(list))
    ϕ = attr.ib(default=attr.Factory(list))
    ϕdot = attr.ib(default=attr.Factory(list))
    problems = attr.ib(default=attr.Factory(list))

    def _finalise(self):
        """
        Finalise data for storage in hdf5 files
        """
        self.derivs = asarray(self.derivs)
        if self.derivs.size == 0:
            self.derivs.shape = (0, len(ODEIndex))

        self.params = asarray(self.params)
        if self.params.size == 0:
            self.params.shape = (0, len(ODEIndex))

        self.times = asarray(self.times)
        self.a1 = asarray(self.a1)
        self.a1dot = asarray(self.a1dot)
        self.a2 = asarray(self.a2)
        self.a2dot = asarray(self.a2dot)
        self.a3 = asarray(self.a3)
        self.a3dot = asarray(self.a3dot)
        self.θ = asarray(self.θ)
        self.θdot = asarray(self.θdot)
        self.ϕdot = asarray(self.ϕdot)

    def __add__(self, other):
        """
        This is required for adding two internal data sets together
        """
        self._finalise()
        other._finalise()

        return InternalData(
            derivs=concatenate((self.derivs, other.derivs)),
            params=concatenate((self.params, other.params)),
            times=concatenate((self.times, other.times)),
            a1=concatenate((self.a1, other.a1)),
            a1dot=concatenate((self.a1dot, other.a1dot)),
            a2=concatenate((self.a1, other.a1)),
            a2dot=concatenate((self.a1dot, other.a1dot)),
            a3=concatenate((self.a1, other.a1)),
            a3dot=concatenate((self.a1dot, other.a1dot)),
            ϕdot=concatenate((self.ϕdot, other.ϕdot)),
            problems=self.problems + other.problems,
        )


@attr.s
class SolutionPhysicalUnits:
    """
    Datastructure of solutions with physical units
    """
    times = attr.ib()
    solution = attr.ib()
    ρ_real = attr.ib()
    ρ_tides = attr.ib()
    ρ_pressure = attr.ib()
    ρ_real_over_ρ_tides = attr.ib()
    ρ_pressure_over_ρ_tides = attr.ib()
    ρ_real_over_ρ_pressure = attr.ib()
    mass = attr.ib()
    len_scale = attr.ib()
    time_scale = attr.ib()
    mass_scale = attr.ib()



"""
File which defines the ODE system and creates a callable solver.
All work is done in dimensionless units unless stated
"""
from os import mkdir
from os.path import isdir
from copy import deepcopy
from numpy import copy, linspace, concatenate, asarray, append as npappend
from mpmath import mp

from scikits.odes import ode
from scikits.odes.sundials import CVODESolveFailed, CVODESolveFoundRoot, CVODESolveReachedTSTOP
from scikits.odes.sundials.cvode import StatusEnum

from src.utils import ODEIndex, EllipIndex, InitConIndex
from src.data_formats import InternalData, Solution, create_results_file
from src.solve.utils import xy_close_stop_generator, get_jump_amount, ontstop_cont, ontstop_stop
from src.solve.geometry import get_Ax, get_Ay, get_Az
from src.solve.deriv_funcs import (
    deriv_x_func, deriv_xdot_func, deriv_y_func, deriv_ydot_func, deriv_z_func, deriv_zdot_func, deriv_θ_func,
    deriv_θdot_func, deriv_ϕ_func, deriv_ϕdot_func, deriv_ϕθdot_func
)


def ode_system(*, a1, a2, a3, ρ_pressure_over_ρ_tides, ρ_real_over_ρ_pressure, store_internal=True):
    """
    Create the mathematical ODE model for the tri-axial ellipsoid.
    Returns the sundials ode model which will be solved.
    """
    # Constants of equations
    ρ_real_over_ρ_pressure = ρ_real_over_ρ_pressure * (a1 * a2 * a3)

    if store_internal:
        internal_data = InternalData()
        times_list = internal_data.times
        params_list = internal_data.params
        derivs_list = internal_data.derivs
        a1_list = internal_data.a1
        a2_list = internal_data.a2
        a3_list = internal_data.a3
        θ_list = internal_data.θ
        ϕ_list = internal_data.ϕ
        a1dot_list = internal_data.a1dot
        a2dot_list = internal_data.a2dot
        a3dot_list = internal_data.a3dot
        θdot_list = internal_data.θdot
        ϕdot_list = internal_data.ϕdot
        problems = internal_data.problems
    else:
        internal_data = None

    def rhs_equation(time, params, derivs):
        """
        RHS of the ODE equations
        Inputs are passed by Sundials.
        """
        # Specify the variables
        x = params[ODEIndex.x]
        y = params[ODEIndex.y]
        z = params[ODEIndex.z]
        θ = params[ODEIndex.θ]
        ϕ = params[ODEIndex.ϕ]
        xdot = params[ODEIndex.xdot]
        ydot = params[ODEIndex.ydot]
        zdot = params[ODEIndex.zdot]
        θdot = params[ODEIndex.θdot]
        ϕdot = params[ODEIndex.ϕdot]

        # Symbols and time-dependent-constants for equations
        Ai = [get_Ax(x=x, y=y, z=z), get_Ay(x=x, y=y, z=z), get_Az(x=x, y=y, z=z)]
        ρ_real_over_ρ_pressure_updating = ρ_real_over_ρ_pressure/(x * y * z)
        #
        # print("Ai", Ai)
        # print("A1*x*y*z", Ai[0] * x * y * z * ρ_real_over_ρ_pressure_updating)
        # print("A1*mass_r", Ai[0] * 0.6987712429686843)
        # print("A2*x*y*z", Ai[1] * x * y * z * ρ_real_over_ρ_pressure_updating)
        # print("A2*mass", Ai[1] * 0.6987712429686843)
        # print("A3*x*y*z", Ai[2] * x * y * z * ρ_real_over_ρ_pressure_updating)
        # print("A3*mass", Ai[0] * 0.6987712429686843)
        # Physical checks:
        if x <= 0.01 or y <= 0.01 or z <= 0.01:
            if store_internal:
                problems.append("-1. Lengths about to negative, lengths were {}, {}, {}".format(x, y, z))
                print("Lengths about to go negative, they are {}. Stopping.".format(x, y, z))
            return -1

        # First order derivative functions
        deriv_x, x_i = deriv_x_func(xdot=xdot)
        deriv_y, y_i = deriv_y_func(ydot=ydot)
        deriv_z, z_i = deriv_z_func(zdot=zdot)
        deriv_θ, θ_i = deriv_θ_func(θdot=θdot)
        deriv_ϕ, ϕ_i = deriv_ϕ_func(ϕdot=ϕdot)

        # Second order derivative functions
        deriv_xdot, xdot_i = deriv_xdot_func(x=x, y=y, z=z, θ=θ, θdot=θdot, ϕdot=ϕdot-θdot,
                                             A1=Ai[EllipIndex.x],
                                             ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure_updating,
                                             ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)
        deriv_ydot, ydot_i = deriv_ydot_func(x=x, y=y, z=z, θ=θ, θdot=θdot, ϕdot=ϕdot-θdot,
                                             A2=Ai[EllipIndex.y],
                                             ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure_updating,
                                             ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)
        deriv_zdot, zdot_i = deriv_zdot_func(x=x, y=y, z=z, A3=Ai[EllipIndex.z],
                                             ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure_updating,
                                             ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)
        deriv_θdot, θdot_i = deriv_θdot_func(x=x, xdot=xdot, y=y, ydot=ydot, θ=θ, θdot=θdot, ϕdot=ϕdot-θdot,
                                             ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)
        deriv_ϕdot, ϕdot_i = deriv_ϕθdot_func(x=x, xdot=xdot, y=y, ydot=ydot, θ=θ, θdot=θdot, ϕdot=ϕdot-θdot,
                                              ρ_pressure_over_ρ_tides=ρ_pressure_over_ρ_tides)

        # Solve
        derivs[ODEIndex.x] = deriv_x
        derivs[ODEIndex.y] = deriv_y
        derivs[ODEIndex.z] = deriv_z
        derivs[ODEIndex.θ] = deriv_θ
        derivs[ODEIndex.ϕ] = deriv_ϕ
        derivs[ODEIndex.xdot] = deriv_xdot
        derivs[ODEIndex.ydot] = deriv_ydot
        derivs[ODEIndex.zdot] = deriv_zdot
        derivs[ODEIndex.θdot] = deriv_θdot
        derivs[ODEIndex.ϕdot] = deriv_ϕdot

        if store_internal:
            params_list.append(copy(params))
            derivs_list.append(copy(derivs))
            times_list.append(copy(time))
            a1_list.append(copy(x_i))
            a2_list.append(copy(y_i))
            a3_list.append(copy(z_i))
            θ_list.append(copy(θ_i))
            ϕ_list.append(copy(ϕ_i))
            a1dot_list.append(copy(xdot_i))
            a2dot_list.append(copy(ydot_i))
            a3dot_list.append(copy(zdot_i))
            θdot_list.append(copy(deriv_θdot))
            ϕdot_list.append(copy(deriv_ϕdot))

    return rhs_equation, internal_data


def calc_numerical_solution(*, initial_conditions, solver_config, time, tstop):
    """
    Run the sundials ODE solver on the set of differential equations
    """
    # initialise the differential equation model
    system, internal_data = ode_system(
       a1=initial_conditions.ode_init_con[InitConIndex.a1],
       a2=initial_conditions.ode_init_con[InitConIndex.a2],
       a3=initial_conditions.ode_init_con[InitConIndex.a3],
       ρ_pressure_over_ρ_tides=initial_conditions.ρ_pressure_over_ρ_tides,
       ρ_real_over_ρ_pressure=initial_conditions.ρ_real_over_ρ_pressure
       )

    # If taylor jump is selected, then add the additional params to the ode solver to perform the action
    tay_args = {}
    # If solver_config.enable_taylor_jump:
    if solver_config.enable_taylor_jump:
        tay_args["rootfn"] = xy_close_stop_generator(solver_config.taylor_jump_threshold)
        tay_args["nr_rootfns"] = 1

    # If time stop is selected, then set the solver to stop at this timepoint
    ontstop_func = ontstop_stop() if solver_config.enable_tstop else ontstop_cont()

    # Create the sundials system
    solver = ode(
        'cvode', system,
        old_api=False,
        validate_flags=True,
        rtol=solver_config.relative_tolerance,
        atol=solver_config.absolute_tolerance,
        max_steps=solver_config.max_steps,
        tstop=tstop,
        ontstop=ontstop_func,
        **tay_args
    )

    # and solve it.
    try:
        soln = solver.solve(time, initial_conditions.ode_init_con)
    except CVODESolveFailed as e:
        soln = e.soln
        print("Solver: FAILED at time {} with conditions {}".format(soln.values.t[-1], soln.values.y[-1, :]))
    except CVODESolveFoundRoot as e:
        soln = e.soln
        print("Solver: FOUND ROOT at time {} with conditions {}".format(soln.values.t[-1], soln.values.y[-1, :]))
    except CVODESolveReachedTSTOP as e:
        soln = e.soln
        print("Solver: TIME STOP at time {}".format(soln.values.t[-1]))
    # except:  # Bare except for problems with division by zero
    #     raise SystemExit("Something went horribly wrong when trying to solve the ODE. Exiting")

    return soln, internal_data


def calc_xy_singularity(*, current_sol, jump_amount, system_config):
    """
    Preform a taylor series expansion of the position and velocity at the input time.
    """
    # Get the most recent time and list of values.
    continue_time = current_sol.values.t[-1]
    continue_soln = current_sol.values.y[-1]
    # time period to jump over
    dt = jump_amount

    # Split equations into components
    x = continue_soln[ODEIndex.x]
    y = continue_soln[ODEIndex.y]
    z = continue_soln[ODEIndex.z]
    θ = continue_soln[ODEIndex.θ]
    ϕ = continue_soln[ODEIndex.ϕ]
    xdot = continue_soln[ODEIndex.xdot]
    ydot = continue_soln[ODEIndex.ydot]
    zdot = continue_soln[ODEIndex.zdot]
    θdot = continue_soln[ODEIndex.θdot]
    ϕdot = continue_soln[ODEIndex.ϕdot]

    # Obtain other components needed to calculate the derivatives
    Ai = [get_Ax(x=x, y=y, z=z), get_Ay(x=x, y=y, z=z), get_Az(x=x, y=y, z=z)]
    ρ_real_over_ρ_pressure_updating = current_sol.ρ_real_over_ρ_pressure * (x * y * z)

    # Numerically calculate the first and second order derivatives at the time point using the previous timestep values
    deriv_xdot, _ = deriv_xdot_func(x=x, y=y, z=z, θ=θ, θdot=θdot, ϕdot=ϕdot-θdot, A1=Ai[EllipIndex.x],
                                    ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure_updating,
                                    ρ_pressure_over_ρ_tides=current_sol.ρ_pressure_over_ρ_tides)
    deriv_ydot, _ = deriv_ydot_func(x=x, y=y, z=z, θ=θ, θdot=θdot, ϕdot=ϕdot-θdot, A2=Ai[EllipIndex.y],
                                    ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure_updating,
                                    ρ_pressure_over_ρ_tides=current_sol.ρ_pressure_over_ρ_tides)
    deriv_zdot, _ = deriv_zdot_func(x=x, y=y, z=z, A3=Ai[EllipIndex.z],
                                    ρ_real_over_ρ_pressure=ρ_real_over_ρ_pressure_updating,
                                    ρ_pressure_over_ρ_tides=current_sol.ρ_pressure_over_ρ_tides)
    deriv_θdot, _ = deriv_θdot_func(x=x, xdot=xdot, y=y, ydot=ydot, θ=θ, θdot=θdot, ϕdot=ϕdot-θdot,
                                    ρ_pressure_over_ρ_tides=current_sol.ρ_pressure_over_ρ_tides)
    deriv_ϕdot, _ = deriv_ϕθdot_func(x=x, xdot=xdot, y=y, ydot=ydot, θ=θ, θdot=θdot, ϕdot=ϕdot-θdot,
                                     ρ_pressure_over_ρ_tides=current_sol.ρ_pressure_over_ρ_tides)

    # Calculate the values of the jump
    x = x + xdot * dt + 1/2 * deriv_xdot * dt**2
    y = y + ydot * dt + 1/2 * deriv_ydot * dt**2
    z = z + zdot * dt + 1/2 * deriv_zdot * dt**2
    θ = θ + θdot * dt + 1/2 * deriv_θdot * dt**2
    ϕ = ϕ + ϕdot * dt + 1/2 * deriv_ϕdot * dt**2
    xdot = xdot + deriv_xdot * dt
    ydot = ydot + deriv_ydot * dt
    zdot = zdot + deriv_zdot * dt
    θdot = θdot + deriv_θdot * dt
    ϕdot = ϕdot + deriv_ϕdot * dt

    return continue_time + dt, [x, xdot, y, ydot, z, zdot, θ, θdot, ϕ, ϕdot]


def solution(*, initial_conditions, solver_config, save_data, folder_name):
    """
    Compute the solution to the model given the input initial conditions and solver configuration
    Saves the solution file.
    :return: List of lists: solution, times, internal_data object
    """
    mp.dps = 50

    # Initialise the time over which to solve the system
    time = linspace(solver_config.start, solver_config.stop, int(solver_config.num_time))
    time_length_passed = solver_config.num_time

    # Compute the solution
    if not solver_config.enable_taylor_jump and not solver_config.enable_tstop:  # No Taylor jump or tstop
        print("Solving the ODE system!")
        soln, internal_data = calc_numerical_solution(
            initial_conditions=initial_conditions,
            solver_config=solver_config,
            time=time,
            tstop=0)
        soln_y = soln.values.y
        soln_t = soln.values.t
        soln_flag = [soln.flag]
        jump_times = []

    elif solver_config.enable_taylor_jump or solver_config.enable_tstop:  # Either Taylor jumps or tstops

        # Checks to make sure everything required for either tay jumps or tstops is present.
        if solver_config.enable_tstop:
            if len(initial_conditions.after_tstop_params) != len(solver_config.tstop_times):
                raise RuntimeError("Input Error: Amount of tstops ({}) != amount of new params ({})".format(
                    initial_conditions.after_tstop_params, solver_config.tstop_times))
            if solver_config.tstop_times[-1] > solver_config.stop:
                raise RuntimeError("Input Error: Selected to use tstops after simulation will end")
        if solver_config.enable_taylor_jump:
            if solver_config.taylor_jump_threshold <= 0:
                raise RuntimeError("Input Error: Selected non-positive ({}) taylor jump threshold".format(
                    solver_config.taylor_jump_threshold))

        # Initialise the data objects to solve the solution where we can jump several times
        soln_y = None
        soln_t = None
        soln_flag = None
        internal_data = None
        jump_times = []

        # Set the first round of initial conditions for the ode
        current_soln_initial_conditions = deepcopy(initial_conditions)

        # Create the tstops and taylor jumps
        tstop_times_remaining = solver_config.tstop_times.copy()
        tstop_times_remaining.append(0)  # Add a zero so that the final run will not stop until specified end time
        tay_jumps_remaining = solver_config.taylor_jumps_num

        print("Solving the ODE system!")
        while True:

            # Compute solution
            current_soln, current_internal_data = calc_numerical_solution(
                initial_conditions=current_soln_initial_conditions,
                solver_config=solver_config,
                time=time,
                tstop=tstop_times_remaining[0])

            # save solution
            if soln_y is None:
                soln_y = current_soln.values.y
                soln_t = current_soln.values.t
                soln_flag = [current_soln.flag]
                internal_data = current_internal_data

            else:
                soln_y = concatenate((soln_y, current_soln.values.y))
                soln_t = concatenate((soln_t, current_soln.values.t))
                soln_flag = soln_flag + [current_soln.flag]
                internal_data = internal_data + current_internal_data

            # Check why the solver stopped.
            if current_soln.flag == StatusEnum.SUCCESS:  # Solver Completed
                print("Solver succeeded.")
                break

            elif current_soln.flag == StatusEnum.ROOT_RETURN:  # Solver found root, i.e. theta double dot singularity
                print("Solver: attempting to preform Taylor jump procedure.")
                if not solver_config.enable_taylor_jump:
                    print("Solver stopped because it found the root but Taylor jump is disabled")
                    break
                else:
                    if tay_jumps_remaining != 0:
                        print("Performing taylor jump at {}".format(current_soln.values.t[-1]))
                        jump_amount = get_jump_amount(
                            x=current_soln.values.y[-1, ODEIndex.x],
                            xdot=current_soln.values.y[-1, ODEIndex.xdot],
                            y=current_soln.values.y[-1, ODEIndex.y],
                            ydot=current_soln.values.y[-1, ODEIndex.ydot],
                            override=False  # can override with manual number. e.g. 3e-3
                        )
                        post_jump_time, post_jump_sol = calc_xy_singularity(
                            current_sol=current_soln,
                            jump_amount=jump_amount,
                            system_config=initial_conditions
                        )
                        # Save the jump data
                        soln_y = concatenate((soln_y, asarray([post_jump_sol])), axis=0)
                        soln_t = npappend(soln_t, post_jump_time)
                        soln_flag = soln_flag + [current_soln.flag]
                        jump_times.append("Solver: Successfully performed Taylor jump at time {}".format(
                            current_soln.values.t[-1]))
                        tay_jumps_remaining -= 1

                        # Check if we taylor jumped over the specified stopping time
                        if post_jump_time > solver_config.stop:
                            print("Solver: Jumped over the wanted stop time")
                            break
                        # Check if Taylor series jumped over a tstop time.
                        if solver_config.enable_tstop:
                            if tstop_times_remaining != [0]:
                                if post_jump_sol > tstop_times_remaining[0]:
                                    print("Solver: Taylor series jumped over a tstop time. I will continue anyway.")
                        # Reset the initial conditions of the solver to the post-jump conditions
                        current_soln_initial_conditions.ode_init_con = post_jump_sol
                        time_length_passed = time_length_passed - len(current_soln.values.t)
                        time = linspace(post_jump_time, solver_config.stop, time_length_passed)

                    else:
                        print("The solver wanted to preform a Taylor Jump but you run out of allowed Jumps. "
                              "Try increasing the amount of allowed jumps")
                        break

            elif current_soln.flag == StatusEnum.TSTOP_RETURN:  # Solver got to the required stop time.
                print("Solver: attempting to preform tstop procedure.")
                if not solver_config.enable_tstop:
                    print("Solver wanted to stop because it reached tstop but tstop is disabled.")
                # Change the initial conditions for the new run
                current_soln_initial_conditions.ode_init_con = current_soln.values.y[-1, :]
                time_length_passed = time_length_passed - len(current_soln.values.t)
                time = linspace(current_soln.values.t[-1], solver_config.stop, time_length_passed)
                # Update the initial conditions that the solver will call after tstop.
                print("vals going in are", current_soln_initial_conditions)
                for key, val in current_soln_initial_conditions.after_tstop_params[0].items():
                    # Set the new initial conditions to be the updated initial conditions
                    print("key, val", key, val)
                    print("inital ones ", getattr(initial_conditions, key))
                    print("thing before", getattr(current_soln_initial_conditions, key))
                    if key == "ode_init_con":
                        setattr(current_soln_initial_conditions, key, val * getattr(current_soln_initial_conditions, key))
                    else:
                        setattr(current_soln_initial_conditions, key, val * getattr(initial_conditions, key))
                    print("thing after", getattr(current_soln_initial_conditions, key))
                current_soln_initial_conditions.after_tstop_params.pop(0)
                # Change the tstop for the following run.
                if tstop_times_remaining == [0]:
                    print("Solver has completed all the required tstops. Will now compute the solution to end of time")
                else:
                    tstop_times_remaining.pop(0)
                    print("Solver: preformed the tstop procedure")

            elif current_soln.flag == StatusEnum.WARNING:
                print("The solver completed solving the solution but something unusable happened")
                print("The message was {}".format(current_soln.message))

            elif current_soln.flag < 0:
                print("The solver failed.")
                print("Flag = {}".format(current_soln.flag))
                print("Message = {}".format(current_soln.message))
                break

            else:
                raise SystemExit("The solver returned a unknown flag. Exiting")

    else:
        print("Could not figure out how you wanted to solve the system.")
        raise SystemExit

    # Create and save the results file
    if save_data:
        if not isdir(folder_name):
            print("No directory '{}' existed. Creating directory.".format(folder_name))
            try:
                mkdir(folder_name)
            except OSError:
                print("Could not create the directory {}.".format(folder_name))
                print("Fix directory or set save data to False")
                raise SystemExit
            else:
                print("Directory created")

        create_results_file(
            Solution(solution=soln_y, times=soln_t, flags=soln_flag, tstop_times=solver_config.tstop_times,
                     jump_times=jump_times),
            internal_data,
            initial_conditions,
            solver_config,
            folder_name,
        )

    return soln_y, soln_t, soln_flag, internal_data



"""
Constants that will be required for the models
"""

GRAV = 6.67408e-8                 # Gravitational constant in CGS
BOLTZMANN = 1.38064852e-16            # Boltzmann
H1MASS = 1.6735328363560003e-24    # Mass of a Hydrogen atom in grams
PARSEC = 3.08567758123e18      # 1 parsec in cm
SOLARMASS = 1.989e33           # Solar mass in grams
YEAR = 60 * 60 * 24 * 365      # 1 year in seconds



"""
prepares the data classes to be dumped to hdf5 files
"""

import h5py
import arrow
from numpy import array


def create_results_file(solution_dump, internal_data_dump, init_con_dump, solver_config_dump, folder_name):
    """
    Save the results from a solution run in a hdf5 file
    inputs are the 4 data objects/classes
    """
    # todo: far future: implement naming convention based on system parameters not time solved at.
    if len(solution_dump.times) > 1:  # Require a solution exists.
        # with h5py.File("{}/TidalStability_".format(folder_name) + "sg_tides_{}_sg_p_{}_mr_{}".format(
        #     init_con_dump.ρ_real_over_ρ_tides, init_con_dump.ρ_real_over_ρ_pressure,
        #     init_con_dump.mass_r * 1/(5/16 * sqrt(5))) + ".hdf5", "x") as f:
        with h5py.File("{}/TidalStability_".format(folder_name) + str(arrow.now().format("YYYY-MM-DD--HH-mm")), "x") as f:
            _solution_dumper(f.create_group("solution"), solution_dump, solver_config_dump)
            _initial_conditions(f.create_group("initial_conditions"), init_con_dump, solver_config_dump)
            _config_dumper(f.create_group("solver_config"), solver_config_dump)
            _internal_data_dumper(f.create_group("internal_data"), internal_data_dump)
            f.attrs["date_solved"] = [str(arrow.now().format("YYYY-MM-DD--HH-mm"))]
            # print("Solution saved as '/{}/TidalStability_".format(folder_name)
            #       + "sg_tides_{}_sg_p_{}_mr_{}".format(init_con_dump.ρ_real_over_ρ_tides,
            #       init_con_dump.ρ_real_over_ρ_pressure, init_con_dump.mass_r * 1/(5/16 * sqrt(5))) + ".hdf5'")
            print("Solution saved as '/{}/TidalStability_".format(folder_name)
                  + str(arrow.now().format("YYYY-MM-DD--HH-mm")) + ".hdf5")
            # Doesn't return anything, just saves the file.
    else:
        print("No properly formatted solution was given. No hdf5 file created.")


def _solution_dumper(grp, solution, solver_config):
    """
    Solution to the ODE dumper
    :param: grp: hdf5 group
    :param: solution: A solution data class instance i.e. Solution class
    """
    grp.create_dataset("times", data=solution.times)
    grp.create_dataset("solution", data=solution.solution)
    grp.create_dataset("flags", data=solution.flags)
    if solver_config.enable_tstop:
        try:
            grp.attrs["tstop_times"] = [time for time in solution.tstop_times]
        except TypeError:
            grp.attrs["tstop_times"] = ["No_tstops"]
    else:
        grp.attrs["tstop_times"] = ["NOT_ENABLED"]
    if solver_config.enable_taylor_jump:
        try:
            grp.attrs["jump_times"] = [i.encode('utf8') for i in solution.jump_times]
        except TypeError:
            grp.attrs["jump_times"] = ["No_Jumps"]
    else:
        grp.attrs["jump_times"] = ["NOT_ENABLED"]
    return grp


def _initial_conditions(grp, init_con, solver_config):
    """
    Initial conditions dumper
    """

    grp.create_dataset("ode_initial_conditions", data=init_con.ode_init_con)
    grp.attrs["ρ_real_over_ρ_tides"] = init_con.ρ_real_over_ρ_tides
    grp.attrs["ρ_pressure_over_ρ_tides"] = init_con.ρ_pressure_over_ρ_tides
    grp.attrs["ρ_real_over_ρ_pressure"] = init_con.ρ_real_over_ρ_pressure
    grp.attrs["mass_r"] = init_con.mass_r
    grp.attrs["equ_radius"] = init_con.equ_radius
    if solver_config.enable_tstop:
        grp.attrs["after_tstop_params"] = str(init_con.after_tstop_params)
    else:
        grp.attrs["after_tstop_params"] = str([])

    return grp


def _config_dumper(grp, config):
    """
    Configuration dumper
    """
    grp.attrs["start"] = config.start
    grp.attrs["stop"] = config.stop
    grp.attrs["num_time"] = config.num_time
    grp.attrs["max_steps"] = config.max_steps
    grp.attrs["relative_tolerance"] = config.relative_tolerance
    grp.attrs["absolute_tolerance"] = config.absolute_tolerance
    grp.attrs["taylor_jump_enabled"] = config.enable_taylor_jump
    grp.attrs["taylor_jumps_num"] = config.taylor_jumps_num
    grp.attrs["taylor_jump_threshold"] = config.taylor_jump_threshold
    grp.attrs["enable_tstop"] = config.enable_tstop
    grp.attrs["tstop_times"] = config.tstop_times
    return grp


def _internal_data_dumper(grp, internal_data):
    """
    Internal data dumper
    """
    grp.create_dataset("params", data=internal_data.params)
    grp.create_dataset("derivs", data=internal_data.derivs)
    grp.create_dataset("times", data=internal_data.times)
    grp.create_dataset("a1", data=internal_data.a1)
    grp.create_dataset("a1dot", data=internal_data.a1dot)
    grp.create_dataset("a2", data=internal_data.a2)
    grp.create_dataset("a2dot", data=internal_data.a2dot)
    grp.create_dataset("a3", data=internal_data.a3)
    grp.create_dataset("a3dot", data=internal_data.a3dot)
    grp.create_dataset("θ", data=internal_data.θ)
    grp.create_dataset("θdot", data=internal_data.θdot)
    grp.create_dataset("ϕ", data=internal_data.ϕ)
    grp.create_dataset("ϕdot", data=array(internal_data.ϕdot))
    grp.attrs["problems"] = [i.encode('utf8') for i in internal_data.problems]
    return grp



# It is with great thanks I mention James Tocknell (Github: Aragilar) for his continued assistance in providing
# solutions to my coding woes
"""
Defines common data structures and how they are to be written to files
"""
from ._containers import InitialConditions, ConfigParams, Solution, InternalData, SolutionPhysicalUnits
from ._dumpers import create_results_file
from ._loaders import solution_loader



import setuptools

version = None
with open('src/tidal_stability/__version__.py', 'r') as f:
    exec(f.read())

setuptools.setup(
    name = "tidal-stability",
    version = version,
    packages = setuptools.find_packages('src'),
    package_dir = {'': 'src'},
    install_requires = [
        "arrow>=0.17.0",
        "attrs>=20.3.0",
        "h5py>=2.10.0",
        "matplotlib>=3.3.2",
        "mpmath>=1.1.0",
        "numpy>=1.19.2",
        "scipy>=1.5.2",
        # todo: sundials -> allows scikits.
    ],
    #python_requires = '>=3.5',
    author = u"Blake Staples",
    author_email = "yourlocalblake@gmail.com",
    description = "Masters thesis code for generating solutions",
    license = "GNU version 3",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3.8',
    ],
    include_package_data=True,
)


"""
Derivatives for the equations of motion
The returned zeros as the second variable of each ODE equation represents a object which can be changed to return any
value wanted for analysis. By returning the full equation in the 2nd spot the individual time steps solved by the
numerical integrator will be returned
"""

from math import sin, cos, sqrt


def deriv_x_func(*, xdot):
    """
    Compute the derivative of the x axis of the ellipsoid
    """
    return xdot, xdot


def deriv_xdot_func(*, x, y, z, θ, θdot, ϕdot, A1, ρ_real_over_ρ_pressure, ρ_pressure_over_ρ_tides):
    """
    Compute the derivative of the derivative of the axis of the ellipsoid
    """
    # x * y * z * ρ_real_over_ρ_pressure  is chosen instead of mass_r as it accounts for the scale factor normalisation
    # that arises from the BE mass ratio deviation from sphericity.
    # They are physically interchangable
    return (
        1/x * (
            + ϕdot * (+ ϕdot * x**2 + 2 * x * y * (+ θdot + 1/sqrt(ρ_pressure_over_ρ_tides)))
            + x**2 * (
                + θdot * (θdot + 2/sqrt(ρ_pressure_over_ρ_tides))
                + 3/ρ_pressure_over_ρ_tides * cos(θ)**2
                - 9/2 * x * y * z * ρ_real_over_ρ_pressure * A1
            )
            + 5 * (1 - 1/ρ_real_over_ρ_pressure)
        )
        , 0
    )


def deriv_y_func(*, ydot):
    """
    Compute the derivative of the y axis of the ellipsoid
    """
    return ydot, ydot


def deriv_ydot_func(*, x, y, z, θ, θdot, ϕdot, A2, ρ_real_over_ρ_pressure, ρ_pressure_over_ρ_tides):
    """
    Compute the derivative of the derivative of the axis of the ellipsoid ellipsoid
    """
    return (
        1/y * (
            + ϕdot * (+ ϕdot * y**2 + 2 * x * y * (+ θdot + 1/sqrt(ρ_pressure_over_ρ_tides)))
            + y**2 * (
                + θdot * (θdot + 2/sqrt(ρ_pressure_over_ρ_tides))
                + 3/ρ_pressure_over_ρ_tides * sin(θ)**2
                - 9/2 * A2 * x * y * z * ρ_real_over_ρ_pressure
            )
            + 5 * (1 - 1/ρ_real_over_ρ_pressure)
        )
        , 0
    )


def deriv_z_func(*, zdot):
    """
    Compute the derivative of the z axis of the ellipsoid
    """
    return zdot, zdot


def deriv_zdot_func(*, x, y, z, A3, ρ_real_over_ρ_pressure, ρ_pressure_over_ρ_tides):
    """
    Compute the derivative of the derivative of the axis of the ellipsoid
    """
    return (
        1/z * (
            - z**2 * (
                + 9/2 * A3 * x * y * z * ρ_real_over_ρ_pressure
                + 1/ρ_pressure_over_ρ_tides
            )
            + 5 * (1 - 1/ρ_real_over_ρ_pressure)
        )
        , 0
    )


def deriv_θ_func(*, θdot):
    """
    Compute the derivative of the θ axis of the ellipsoid
    """
    return θdot, θdot


def deriv_θdot_func(*, x, xdot, y, ydot, θ, θdot, ϕdot, ρ_pressure_over_ρ_tides):
    """
    Compute the derivative of the derivative of the θ axis of the ellipsoid
    """
    return (
        1/((-x + y) * (x + y)) * (
                + 2 * (xdot * x - ydot * y) * (θdot + sqrt(1/ρ_pressure_over_ρ_tides))
                + 2 * ϕdot * (- xdot * y + ydot * x)
                + 3/2 * sin(2 * θ) * (x**2 + y**2) * 1/ρ_pressure_over_ρ_tides
        )
        , 0
    )


def deriv_ϕ_func(*, ϕdot):
    """
    Compute the derivative of the ϕ axis of the ellipsoid
    """
    return ϕdot, ϕdot


def deriv_ϕθdot_func(*,  x, xdot, y, ydot, θ, θdot, ϕdot, ρ_pressure_over_ρ_tides):
    """
    Compute the derivative of the derivative of the ϕ axis of the ellipsoid
    """
    return (
        1/(x + y) * (
            - 2 * (xdot + ydot) * (θdot + ϕdot + sqrt(1/ρ_pressure_over_ρ_tides))
            - 3/2 * sin(2 * θ) * (x - y) * 1/ρ_pressure_over_ρ_tides
        )
        , 0
    )


def deriv_ϕdot_func(*,  x, xdot, y, ydot, θ, θdot, ϕdot, ρ_tides_over_ρ_pressure, sqrt_ρ_tides_over_ρ_pressure):
    """
    Compute the derivative of the derivative of the ϕ axis of the ellipsoid
    """

    return (
        1/((x - y) * (x + y)) * (
            + 2 * ϕdot * (- xdot * x + ydot * y)
            + 2 * (xdot * y - ydot * x) * (θdot + sqrt_ρ_tides_over_ρ_pressure)
            + 3 * ρ_tides_over_ρ_pressure * x * y * sin(2 * θ)
        )
        , 0
    )



"""
Load phy5 files
"""
import h5py

from src.data_formats import Solution, InitialConditions, ConfigParams, InternalData


def solution_loader(file_name=None):
    """
    Load a hdf5 file containing a solved ODE
    Returns the 4 data objects
    """
    if file_name is not None:
        try:
            h5py.File(file_name, "r")
        except OSError:
            try:
                file_name = file_name + ".hdf5"
                h5py.File(file_name, "r")
            except OSError:
                raise SystemExit("Could not find file, got {}".format(file_name))
    else:
        raise SystemExit("No file given, got {}".format(file_name))

    with h5py.File(file_name, "r") as f:
        solution = Solution(
            solution=f["solution"]["solution"][:],
            times=f["solution"]["times"][:],
            flags=f["solution"]["flags"][:],
            tstop_times=f["solution"].attrs["tstop_times"],
            jump_times=f["solution"].attrs["jump_times"]
        )

        config_params = ConfigParams(
            start=f["solver_config"].attrs["start"],
            stop=f["solver_config"].attrs["stop"],
            num_time=f["solver_config"].attrs["num_time"],
            max_steps=f["solver_config"].attrs["max_steps"],
            relative_tolerance=f["solver_config"].attrs["relative_tolerance"],
            absolute_tolerance=f["solver_config"].attrs["absolute_tolerance"],
            enable_taylor_jump=f["solver_config"].attrs["taylor_jump_enabled"],
            taylor_jumps_num=f["solver_config"].attrs["taylor_jumps_num"],
            taylor_jump_threshold=f["solver_config"].attrs["taylor_jump_threshold"],
            enable_tstop=f["solver_config"].attrs["enable_tstop"],
            tstop_times=f["solver_config"].attrs["tstop_times"]
        )

        internal_data = InternalData(
            params=f["internal_data"]["params"][:],
            derivs=f["internal_data"]["derivs"][:],
            times=f["internal_data"]["times"][:],
            a1=f["internal_data"]["a1"][:],
            a1dot=f["internal_data"]["a1dot"][:],
            a2=f["internal_data"]["a2"][:],
            a2dot=f["internal_data"]["a2dot"][:],
            a3=f["internal_data"]["a3"][:],
            a3dot=f["internal_data"]["a3dot"][:],
            θ=f["internal_data"]["θ"][:],
            θdot=f["internal_data"]["θdot"][:],
            ϕ=f["internal_data"]["ϕ"][:],
            ϕdot=f["internal_data"]["ϕdot"][:],
            problems=f["internal_data"].attrs["problems"]
        )

        initial_conditions = InitialConditions(
            ode_init_con=f["initial_conditions"]["ode_initial_conditions"][:],
            ρ_real_over_ρ_tides=f["initial_conditions"].attrs["ρ_real_over_ρ_tides"],
            ρ_pressure_over_ρ_tides=f["initial_conditions"].attrs["ρ_pressure_over_ρ_tides"],
            ρ_real_over_ρ_pressure=f["initial_conditions"].attrs["ρ_real_over_ρ_pressure"],
            mass_r=f["initial_conditions"].attrs["mass_r"],
            equ_radius=f["initial_conditions"].attrs["equ_radius"],
        )

    return solution, initial_conditions, config_params, internal_data
